r"""Implement some conditions that can be used in the event system."""

from __future__ import annotations

__all__ = ["BaseCondition", "PeriodicCondition"]

from abc import ABC, abstractmethod
from typing import Any


class BaseCondition(ABC):
    r"""Define the base class to implement a condition for
    ``ConditionalEventHandler``.

    A child class has to implement the following methods:

        - ``evaluate``
        - ``equal``

    Example usage:

    ```pycon
    >>> from minevent import PeriodicCondition
    >>> condition = PeriodicCondition(freq=3)
    >>> condition.evaluate()
    True
    >>> condition.evaluate()
    False
    >>> condition.evaluate()
    False
    >>> condition.evaluate()
    True
    >>> condition.evaluate()
    False
    >>> condition.evaluate()
    False
    >>> condition.evaluate()
    True

    ```
    """

    def __eq__(self, other: object) -> bool:
        return self.equal(other)

    @abstractmethod
    def equal(self, other: Any) -> bool:
        r"""Compare two conditions.

        Args:
            other: Specifies the other object to compare with.

        Returns:
            ``True`` if the two conditions are equal,
                otherwise ``False``.

        Example usage:

        ```pycon
        >>> from minevent import PeriodicCondition
        >>> condition = PeriodicCondition(freq=3)
        >>> condition.equal(PeriodicCondition(freq=3))
        True
        >>> condition.equal(PeriodicCondition(freq=2))
        False

        ```
        """

    @abstractmethod
    def evaluate(self) -> bool:
        r"""Evaluate the condition given the current state.

        Returns:
            ``True`` if the condition is ``True`` and the event
                handler logic should be executed, otherwise ``False``.

        Example usage:

        ```pycon
        >>> from minevent import EventHandler
        >>> def hello_handler() -> None:
        ...     print("Hello!")
        ...
        >>> handler = EventHandler(hello_handler)
        >>> handler.handle()
        Hello!

        ```
        """


class PeriodicCondition(BaseCondition):
    r"""Implement a periodic condition.

    This condition is true every ``freq`` events.

    Args:
        freq: Specifies the frequency.

    Example usage:

    ```pycon
    >>> from minevent import PeriodicCondition
    >>> condition = PeriodicCondition(freq=3)
    >>> condition.evaluate()
    True
    >>> condition.evaluate()
    False
    >>> condition.evaluate()
    False
    >>> condition.evaluate()
    True
    >>> condition.evaluate()
    False
    >>> condition.evaluate()
    False
    >>> condition.evaluate()
    True

    ```
    """

    def __init__(self, freq: int) -> None:
        self._freq = int(freq)
        self._step = 0

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(freq={self._freq:,}, step={self._step:,})"

    @property
    def freq(self) -> int:
        r"""The frequency of the condition."""
        return self._freq

    def equal(self, other: Any) -> bool:
        if isinstance(other, PeriodicCondition):
            return self.freq == other.freq
        return False

    def evaluate(self) -> bool:
        condition = self._step % self._freq == 0
        self._step += 1
        return condition



r"""Contain the main features of the ``minevent`` package."""

from __future__ import annotations

__all__ = [
    "BaseCondition",
    "BaseEventHandler",
    "BaseEventHandlerWithArguments",
    "ConditionalEventHandler",
    "EventHandler",
    "EventManager",
    "PeriodicCondition",
]


from minevent.conditions import BaseCondition, PeriodicCondition
from minevent.handlers import (
    BaseEventHandler,
    BaseEventHandlerWithArguments,
    ConditionalEventHandler,
    EventHandler,
)
from minevent.manager import EventManager



r"""Implement the event handlers."""

from __future__ import annotations

__all__ = [
    "BaseEventHandler",
    "BaseEventHandlerWithArguments",
    "ConditionalEventHandler",
    "EventHandler",
]

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils import str_indent, str_mapping

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from minevent.conditions import BaseCondition


class BaseEventHandler(ABC):
    r"""Define the base class to implement an event handler.

    A child class has to implement the following methods:

        - ``handle``
        - ``equal``

    Example usage:

    ```pycon
    >>> from minevent import EventHandler
    >>> def hello_handler() -> None:
    ...     print("Hello!")
    ...
    >>> handler = EventHandler(hello_handler)
    >>> handler
    EventHandler(
      (handler): <function hello_handler at 0x...>
      (handler_args): ()
      (handler_kwargs): {}
    )
    >>> handler.handle()
    Hello!

    ```
    """

    def __eq__(self, other: object) -> bool:
        return self.equal(other)

    @abstractmethod
    def equal(self, other: Any) -> bool:
        r"""Compare two event handlers.

        Args:
            other: Specifies the other object to compare with.

        Returns:
            ``True`` if the two event handlers are equal,
                otherwise ``False``.

        Example usage:

        ```pycon
        >>> from minevent import EventHandler
        >>> def hello_handler() -> None:
        ...     print("Hello!")
        ...
        >>> handler = EventHandler(hello_handler)
        >>> handler.equal(EventHandler(hello_handler))
        True
        >>> handler.equal(EventHandler(print, handler_args=["Hello!"]))
        False

        ```
        """

    @abstractmethod
    def handle(self) -> None:
        r"""Handle the event.

        Example usage:

        ```pycon
        >>> from minevent import EventHandler
        >>> def hello_handler() -> None:
        ...     print("Hello!")
        ...
        >>> handler = EventHandler(hello_handler)
        >>> handler.handle()
        Hello!

        ```
        """


class BaseEventHandlerWithArguments(BaseEventHandler):
    r"""Define a base class to implement an event handler with positional
    and/or keyword arguments.

    A child class has to implement the ``equal`` method.

    Args:
        handler: Specifies the handler.
        handler_args: Specifies the positional arguments of the
            handler.
        handler_kwargs: Specifies the arbitrary keyword arguments of
            the handler.

    Example usage:

    ```pycon
    >>> from minevent import EventHandler
    >>> def hello_handler() -> None:
    ...     print("Hello!")
    ...
    >>> handler = EventHandler(hello_handler)
    >>> handler
    EventHandler(
      (handler): <function hello_handler at 0x...>
      (handler_args): ()
      (handler_kwargs): {}
    )
    >>> handler.handle()
    Hello!
    >>> handler = EventHandler(print, handler_args=["Hello!"])
    >>> handler.handle()
    Hello!

    ```
    """

    def __init__(
        self,
        handler: Callable,
        handler_args: Sequence | None = None,
        handler_kwargs: dict | None = None,
    ) -> None:
        if not callable(handler):
            msg = f"handler is not callable: {handler}"
            raise TypeError(msg)
        self._handler = handler
        self._handler_args = tuple(handler_args or ())
        self._handler_kwargs = handler_kwargs or {}

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "handler": self._handler,
                    "handler_args": self._handler_args,
                    "handler_kwargs": self._handler_kwargs,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def handler(self) -> Callable:
        r"""The handler."""
        return self._handler

    @property
    def handler_args(self) -> tuple:
        r"""Variable length argument list of the handler."""
        return self._handler_args

    @property
    def handler_kwargs(self) -> dict:
        r"""Arbitrary keyword arguments of the handler."""
        return self._handler_kwargs

    def handle(self) -> None:
        self._handler(*self._handler_args, **self._handler_kwargs)


class EventHandler(BaseEventHandlerWithArguments):
    r"""Implement a simple event handler.

    Example usage:

    ```pycon
    >>> from minevent import EventHandler
    >>> def hello_handler() -> None:
    ...     print("Hello!")
    ...
    >>> handler = EventHandler(hello_handler)
    >>> handler
    EventHandler(
      (handler): <function hello_handler at 0x...>
      (handler_args): ()
      (handler_kwargs): {}
    )
    >>> handler.handle()
    Hello!

    ```
    """

    def equal(self, other: Any) -> bool:
        if not isinstance(other, EventHandler):
            return False
        return (
            objects_are_equal(self.handler, other.handler)
            and objects_are_equal(self.handler_args, other.handler_args)
            and objects_are_equal(self.handler_kwargs, other.handler_kwargs)
        )


class ConditionalEventHandler(BaseEventHandlerWithArguments):
    r"""Implement a conditional event handler.

    The handler is executed only if the condition is ``True``.

    Args:
        handler: Specifies the handler.
        condition: Specifies the condition for this event handler.
            The condition should be callable without arguments.
        handler_args: Specifies the positional arguments of the
            handler.
        handler_kwargs: Specifies the arbitrary keyword arguments of
            the handler.

    Example usage:

    ```pycon
    >>> from minevent import ConditionalEventHandler, PeriodicCondition
    >>> def hello_handler() -> None:
    ...     print("Hello!")
    ...
    >>> handler = ConditionalEventHandler(hello_handler, PeriodicCondition(freq=3))
    >>> handler
    ConditionalEventHandler(
      (handler): <function hello_handler at 0x...>
      (handler_args): ()
      (handler_kwargs): {}
      (condition): PeriodicCondition(freq=3, step=0)
    )
    >>> handler.handle()
    Hello!
    >>> handler.handle()
    >>> handler.handle()
    >>> handler.handle()
    Hello!

    ```
    """

    def __init__(
        self,
        handler: Callable,
        condition: BaseCondition,
        handler_args: Sequence | None = None,
        handler_kwargs: dict | None = None,
    ) -> None:
        super().__init__(handler=handler, handler_args=handler_args, handler_kwargs=handler_kwargs)
        self._condition = condition

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "handler": self._handler,
                    "handler_args": self._handler_args,
                    "handler_kwargs": self._handler_kwargs,
                    "condition": self._condition,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def condition(self) -> BaseCondition:
        r"""The condition."""
        return self._condition

    def equal(self, other: Any) -> bool:
        if not isinstance(other, ConditionalEventHandler):
            return False
        return (
            objects_are_equal(self.handler, other.handler)
            and objects_are_equal(self.handler_args, other.handler_args)
            and objects_are_equal(self.handler_kwargs, other.handler_kwargs)
            and objects_are_equal(self.condition, other.condition)
        )

    def handle(self) -> None:
        if self._condition.evaluate():
            super().handle()



r"""Implement the event manager."""

from __future__ import annotations

__all__ = ["EventManager"]

import logging
from collections import defaultdict
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping, str_sequence

if TYPE_CHECKING:
    from minevent.handlers import BaseEventHandler

logger = logging.getLogger(__name__)


class EventManager:
    r"""Implement an event manager.

    This event manager allows adding event handlers and firing events.
    An event is represented by a case-sensitive string.

    Example usage:

    ```pycon
    >>> from minevent import EventHandler, EventManager
    >>> def hello_handler():
    ...     print("Hello!")
    ...
    >>> manager = EventManager()
    >>> manager.add_event_handler("my_event", EventHandler(hello_handler))
    >>> manager.trigger_event("my_event")
    Hello!

    ```
    """

    def __init__(self) -> None:
        # This variable is used to store the handlers associated to each event.
        self._event_handlers = defaultdict(list)
        # This variable is used to track the last fired event name
        self._last_triggered_event = None
        self.reset()

    def __repr__(self) -> str:
        event_handlers = str_mapping(
            {
                event: "\n" + str_sequence(handler) if handler else ""
                for event, handler in self._event_handlers.items()
            }
        )
        args = str_indent(
            str_mapping(
                {
                    "event_handlers": "\n" + event_handlers if event_handlers else event_handlers,
                    "last_triggered_event": self._last_triggered_event,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def last_triggered_event(self) -> str | None:
        r"""Get the last event name that was triggered.

        Returns:
            The last event name that was fired of ``None`` if no event
                was fired.

        Example usage:

        ```pycon
        >>> from minevent import EventHandler, EventManager
        >>> manager = EventManager()
        >>> def hello_handler():
        ...     print("Hello!")
        ...
        >>> manager.add_event_handler("my_event", EventHandler(hello_handler))
        >>> manager.trigger_event("my_event")
        Hello!
        >>> manager.last_triggered_event

        ```
        """
        return self._last_triggered_event

    def add_event_handler(self, event: str, event_handler: BaseEventHandler) -> None:
        r"""Add an event handler to an event.

        The event handler will be called everytime the event happens.

        Args:
            event: Specifies the event to attach the event handler.
            event_handler: Specifies the event handler to attach to
                the event.

        Example usage:

        ```pycon
        >>> from minevent import EventManager, EventHandler
        >>> def hello_handler():
        ...     print("Hello!")
        ...
        >>> manager = EventManager()
        >>> manager.add_event_handler("my_event", EventHandler(hello_handler))

        ```
        """
        self._event_handlers[str(event)].append(event_handler)
        logger.debug(f"Added {event_handler} to event {event}")

    def trigger_event(self, event: str) -> None:
        r"""Trigger the handler(s) for the given event.

        Args:
            event: Specifies the event to fire.

        Example usage:

        ```pycon
        >>> from minevent import EventHandler, EventManager
        >>> manager = EventManager()
        >>> manager.trigger_event("my_event")  # do nothing because there is no event handler
        >>> def hello_handler():
        ...     print("Hello!")
        ...
        >>> manager.add_event_handler("my_event", EventHandler(hello_handler))
        >>> manager.trigger_event("my_event")
        Hello!

        ```
        """
        logger.debug(f"Firing {event} event")
        self._last_triggered_event = event
        for event_handler in self._event_handlers[event]:
            event_handler.handle()

    def has_event_handler(self, event_handler: BaseEventHandler, event: str | None = None) -> bool:
        r"""Indicate if a handler is registered in the event manager.

        Note that this method relies on the ``equal`` method of the
        input event handler to compare event handlers.

        Args:
            event_handler: Specifies the event handler to check.
            event: Specifies an event to check. If the value is
                ``None``, it will check all the events.

        Example usage:

        ```pycon
        >>> from minevent import EventHandler, EventManager
        >>> def hello_handler():
        ...     print("Hello!")
        ...
        >>> manager = EventManager()
        >>> # Check if `hello_handler` is registered in the event manager
        >>> manager.has_event_handler(EventHandler(hello_handler))
        False
        >>> # Check if `hello_handler` is registered in the event manager for 'my_event' event
        >>> manager.has_event_handler(EventHandler(hello_handler), "my_event")
        False
        >>> # Add an event handler
        >>> manager.add_event_handler("my_event", EventHandler(hello_handler))
        >>> # Check if `hello_handler` is registered in the event manager
        >>> manager.has_event_handler(EventHandler(hello_handler))
        True
        >>> # Check if `hello_handler` is registered in the event manager for 'my_event' event
        >>> manager.has_event_handler(EventHandler(hello_handler), "my_event")
        True
        >>> # Check if `hello_handler` is registered in the event manager for 'my_other_event' event
        >>> manager.has_event_handler(EventHandler(hello_handler), "my_other_event")
        False

        ```
        """
        events = [event] if event else self._event_handlers
        for evnt in events:
            for handler in self._event_handlers[evnt]:
                if event_handler.equal(handler):
                    return True
        return False

    def remove_event_handler(self, event: str, event_handler: BaseEventHandler) -> None:
        r"""Remove an event handler of a given event.

        Note that if the same event handler was added multiple times
        the event, all the duplicated handlers are removed. This
        method relies on the ``equal`` method of the input event
        handler to compare event handlers.

        Args:
            event: Specifies the event handler is attached to.
            event_handler: Specifies the event handler to remove.

        Raises:
            ValueError: if the event does not exist or if the handler
                is not attached to the event.

        Example usage:

        ```pycon
        >>> from minevent import EventHandler, EventManager
        >>> manager = EventManager()
        >>> def hello_handler():
        ...     print("Hello!")
        ...
        >>> manager.add_event_handler("my_event", EventHandler(hello_handler))
        >>> # Check if `hello_handler` is registered in the event manager for 'my_event' event
        >>> manager.has_event_handler(EventHandler(hello_handler), "my_event")
        True
        >>> # Remove the event handler of the engine
        >>> manager.remove_event_handler("my_event", EventHandler(hello_handler))
        >>> # Check if `hello_handler` is registered in the event manager for 'my_event' event
        >>> manager.has_event_handler(EventHandler(hello_handler), "my_event")
        False

        ```
        """
        if event not in self._event_handlers:
            msg = f"'{event}' event does not exist"
            raise RuntimeError(msg)

        new_event_handlers = [
            handler for handler in self._event_handlers[event] if not event_handler.equal(handler)
        ]
        if len(new_event_handlers) == len(self._event_handlers[event]):
            msg = (
                f"{event_handler} is not found among registered event handlers for '{event}' event"
            )
            raise RuntimeError(msg)
        if len(new_event_handlers) > 0:
            self._event_handlers[event] = new_event_handlers
        else:
            del self._event_handlers[event]
        logger.debug(f"Removed {event_handler} in '{event}' event")

    def reset(self) -> None:
        r"""Reset the event manager.

        This method removes all the event handlers from the event manager.

        Example usage:

        ```pycon
        >>> # Create an event manager
        >>> from minevent import EventManager
        >>> manager = EventManager()
        >>> # Add an event handler to the engine
        >>> def hello_handler():
        ...     print("Hello!")
        ...
        >>> from minevent import EventHandler
        >>> manager.add_event_handler("my_event", EventHandler(hello_handler))
        >>> # Check if `hello_handler` is registered in the event manager for 'my_event' event
        >>> manager.has_event_handler(EventHandler(hello_handler), "my_event")
        True
        >>> manager.trigger_event("my_event")
        >>> manager.last_triggered_event
        my_event
        >>> # Reset the event manager
        >>> manager.reset()
        >>> # Check if `hello_handler` is registered in the event manager for 'my_event' event
        >>> manager.has_event_handler(EventHandler(hello_handler), "my_event")
        False
        >>> manager.last_triggered_event
        None

        ```
        """
        self._event_handlers.clear()
        self._last_triggered_event = None
